#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>

#define N 256

/* Gestore del segnale SIGCHLD */
void handler(int s)
{
	int status;
	while (waitpid(-1, &status, WNOHANG) > 0);
}

int main(int argc, char** argv)
{
	struct addrinfo hints, *res;
	int err, sd, ns, pid, pid2, on, piped[2];
	char regione[N], impianti[N];
	// IMPORTANTE: creando l'ack aggiungo \n per consentire al client Java la readLine
	char *ack = "ack\n";

	/* Controllo argomenti */
	if(argc < 2){
		printf("Uso: ./server <porta> \n");
		exit(1);
	}

	memset(&hints, 0, sizeof(hints));
	hints.ai_family   = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags    = AI_PASSIVE;

	if ((err = getaddrinfo(NULL, argv[1], &hints, &res)) != 0) {
		fprintf(stderr, "Errore setup indirizzo bind: %s\n", gai_strerror(err));
		exit(2);
	}

	if ((sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol)) < 0){
		perror("Errore in socket");
		exit(3);
	}

	on = 1;
	if(setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))<0){
		perror("setsockopt");
		exit(4);
	}

	if (bind(sd, res->ai_addr, res->ai_addrlen) < 0) {
		perror("Errore in bind");
		exit(5);
	}

	freeaddrinfo(res);

	if(listen(sd, SOMAXCONN)<0){
		perror("listen");
		exit(6);
	}

	signal(SIGCHLD, handler);

	/* Attendo i client... */
	for(;;){
		printf("Server in ascolto...\n");

		if((ns = accept(sd, NULL, NULL)) < 0) {
			/* Ignoro gli errori di tipo interruzione da segnale. */
			if (errno == EINTR)
				continue;
			/* Gestisco tutte le altre tipologie di errore. */
			perror("accept");
			exit(7);
		}

		/* Generazione di un figlio */
		if((pid = fork())<0){
			perror("fork");
			exit(8);
		} else if(pid == 0){

			/* Chiudo la socket passiva */
			close(sd);

			/* figlio */

			memset(regione, 0, sizeof(regione));
			if(read(ns, regione, N-1)<0){
				perror("read regione");
				exit(9);
			}

			if(write(ns, ack, strlen(ack))<0){
				perror("write ack");
				exit(10);
			}

			memset(impianti, 0, sizeof(impianti));
			if(read(ns, impianti, N-1)<0){
				perror("read impianti");
				exit(10);
			}

			if (pipe(piped) < 0) {
				perror("pipe");
				exit(11);
			}

			if((pid2 = fork())<0){
				perror("seconda fork");
				exit(12);
			}
			else if(pid2 == 0){

				// Nipote N1
				// Chiudo i descrittori che non servono
				close(piped[0]);
				close(ns);

				// Ridireziono stdout
				close(1);
				dup(piped[1]);
				close(piped[1]);

				// Mi metto nel path richiesto
				chdir("./bollettino_neve/");
				// Aggiungo l'estensione file al nome della regione
				strcat(regione, ".txt");

				// Eseguo la sort mandando i risultati al figlio

				execlp("sort", "sort", "-n" ,"-r", regione, (char*)0);
				perror("exec sort");
				exit(13);

			}

			// Figlio

			// Chiudo descrittori non necessari
			close(piped[1]);

			// Ridireziono stdin
			close(0);
			dup(piped[0]);
			close(piped[0]);

			/* Ridireziono stdout */
			close(1);
			dup(ns);
			close(ns);

			execlp("head", "head", "-n", impianti, (char *)0);
			perror("exec head");
			exit(14);

		}
		// padre
		close(ns);
	}
	return 0;
}
